RED = {
    'bg_color': '#FFC7CE',
    'font_color': '#9C0006'
}

YELLOW = {
    'bg_color': '#FFEB9C',
    'font_color': '#9C6500'
}

GREEN = {
    'bg_color': '#C6EFCE',
    'font_color': '#006100'
}

EMPTY_CELL = ""
